#! /bin/awk

# Get along with quotes in the comment string
function format_to_bash_string(_in)
{

	count = split(dos2unix(_in),_out_array,"'");
	_out="'" _out_array[1] "'";
	for(i=2; i <= count; ++i)
	{
		_out = _out "\"'\"'" _out_array[i] "'";
	}
	return _out;
}

function push_cvs_commit(){
	if(debug)
	{
		print "push_cvs_commit:" line_number": ...";
		if(debug_trace)
			print "... `" file_name "' `" flatten(comment_under_construction) "'";
	}
	if(!(file_name in file_array))
	{
		file_array[file_name] = cvs_commit_count;

		if(debug_trace)
			print "... new commit, number=" cvs_commit_count

		cvs_commit_target[cvs_commit_count] = " " file_name;
		comment = format_to_bash_string(comment_under_construction);
		cvs_commit[cvs_commit_count++] = "cvs commit -m " comment;

		if(debug_trace)
			print "... set begin: " cvs_commit[cvs_commit_count-1] cvs_commit_target[cvs_commit_count-1] "\n... set end."


		if(match(comment,"^[ \t]*['\"]?[ \t]*(\\([^)]*\\))?[ \t]*(<[^>]*>)?[ \t]*\\*\\*[ \t]*[Aa](dd(ed|ition)|jout?)"))
		{
			cvs_add[cvs_add_count++] = "cvs add -m " comment " " file_name;
		
		}
		else if(match(comment,"^[ \t]*['\"]?[ \t]*(\\([^)]*\\))?[ \t]*(<[^>]*>)?[ \t]*\\*\\*[ \t]*([Rr]emoved|Suppr(im|ession))"))
		{
			cvs_remove[cvs_remove_count++] = "cvs remove " file_name;
		
		}
		# else there is nothing to do, this is neither an add nor a remove Changelog entry
	}
	else
	{
		i = file_array[file_name];

		if(debug_trace)
			print "... existing commit, number=" i

		comment = format_to_bash_string(comment_under_construction);
		cvs_commit[i] = cvs_commit[i] "\"\n\"" comment;
		if(debug_trace)
			print "... update begin: " cvs_commit[i] cvs_commit_target[i] "\n... update end."

		
	}
}

function rtrim(_in){
	if(match(_in,"[ \t]+$"))
	{
		_in = substr(_in,1,length(_in)-RLENGTH);
	}
	return _in;
}

function dos2unix(_in){
	count= split(_in,_out_array,"\r");
	_out = _out_array[1];
	for(i=2; i < count; ++i)
	{
		_out=_out   _out_array[i] ;
	}

	return _out;
}

# used for debug only
function flatten(_in){
	count= split(dos2unix(_in),_out_array,"\n");
	_out = _out_array[1];
	for(i=2; i < count; ++i)
	{
		_out=_out "^J" _out_array[i] ;
	}
	return _out;
}


BEGIN { 
   comment_under_construction = "";
   if(p ~ "^[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]$")
	   stop_time_stamp="<" p "> ";
   else if(p != "")
   {
	   print "ERROR: invalid p in command line";
	   exit(-1);
   }
   else
	   stop_time_stamp="<0000-01-01> ";

   latest_time_stamp="";
   time_stamp = "";
   time_stamp_line = "";
   line_number = 1;
   cvs_commit_count = 0;
   cvs_add_count = 0;
   cvs_remove_count = 0;
   done = 0;
   debug = 1;
   debug_trace = 1;
}

# time stamp
/^[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]  / && !done { 
	if(latest_time_stamp == "")
		latest_time_stamp=$1
	time_stamp="<" $1 "> "; time_stamp_line=$0; 

	if(time_stamp <= stop_time_stamp)
		nextfile;

	done = 1;
	if(debug){
		print "timestamp:" line_number ": " $0
	}
}

# entry
/^\t/ && !done { 
	line = substr($0, 2);
	if(comment_under_construction == "" && substr(line,1,1) != "*")
	{
		cvs_commit[cvs_commit_count] = "#ERROR:" line_number ": ";
		cvs_commit_target[cvs_commit_count++] = "no interline before time-stamp";
	}
	if(substr(line,1,1) == "(")
	{
		comment_under_construction = comment_under_construction "\n" line;
	}
	else if(comment_under_construction == "")
	{
		comment_under_construction = line;
	}
	else
	{
		comment_under_construction = comment_under_construction " " line;
	}
	done = 1;
	if(debug){
		print "entry:" line_number ": " $0
	}
}

# empty line with some comment already gathered 
($0 ==  "" || $0 ==  "\r") && comment_under_construction != "" { 
	if(substr(comment_under_construction,1,2) != "* "){
		if(debug){
			print "#ERROR:" line_number ": entry not starting by \"* \": ..."
			if(debug_trace) 
				print "... `" flatten(comment_under_construction) "'";
		}
		cvs_commit[cvs_commit_count] = "#ERROR:" line_number ": ";
		cvs_commit_target[cvs_commit_count++] = "entry not starting by \"* \": " format_to_bash_string(flatten(comment_under_construction));
	}
	else if(match(comment_under_construction,"^\\* [^(:]+:"))
	{
		if(debug_trace) 
			print "parsed:" line_number ": `" flatten(comment_under_construction) "'";
		file_name = gensub("^\\* ([^(:]+):.*$","\\1",1,comment_under_construction);
		comment_under_construction = time_stamp gensub("^\\* [^(:]+: *(.*)$","\\1",1,comment_under_construction);
		if(debug_trace) 
		{
			print "file_name:" line_number ": `" flatten(file_name) "'";
			print "comment:" line_number ": `" flatten(comment_under_construction) "'";
		}

		push_cvs_commit();
	}
	else if(match(comment_under_construction,"^\\* [^(:]+ +\\([^)]+\\):"))
	{
		file_name = gensub("^\\* ([^(:]+).*$","\\1",1,comment_under_construction);
		comment_under_construction = time_stamp gensub("^\\* [^(:]+ +(\\([^)]+\\)): *(.*)$" ,"\\1 \\2",1,comment_under_construction);
		push_cvs_commit();
	}
	else
	{
		if(debug){
			print "#ERROR:" line_number ": invalid entry: ..."
			if(debug_trace)
				"...`" flatten(comment_under_construction) "'";
		}
		cvs_commit[cvs_commit_count] = "#ERROR:" line_number ": ";
		cvs_commit_target[cvs_commit_count++] = "invalid entry: " flatten(comment_under_construction);
	}

	comment_under_construction = "";
	done = 1;
	if(debug){
		print "empty line:" line_number ": " $0
	}

}


# empty pattern
    { 
		++line_number;
		done = 0;
	}

# output of script
END {
	print "OK=1;"
	print "# CVS removals"
	for(i=0; i < cvs_remove_count; ++i)
	{
			print cvs_remove[i];
			print "if [ $? -ne 0 ]; then OK=0; fi"
	}
	print "# CVS additions"
	for(i=0; i < cvs_add_count; ++i)
	{
			print cvs_add[i];
			print "if [ $? -ne 0 ]; then OK=0; fi"
	}
	print "# CVS commitments"
	for(i=0; i < cvs_commit_count; ++i)
	{
		print cvs_commit[i] cvs_commit_target[i];
		print "if [ $? -ne 0 ]; then OK=0; fi"

	}
	print "if [ ${OK} -eq 1 ]; then echo \"" latest_time_stamp "\" > ChangeLog_to_cvs_script.latest_timestamp; fi"
}

